-- Copyright 2024 DJ WoodZ

-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:

-- The above copyright notice and this permission notice shall be included in all
-- copies or substantial portions of the Software.

-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.

EnhancedEconomySettings = {}
EnhancedEconomySettings.name = g_currentModName
EnhancedEconomySettings.settingsDir = g_currentModSettingsDirectory

source(g_currentModDirectory .. 'scripts/EnhancedEconomySettingsGUI.lua')
source(g_currentModDirectory .. 'scripts/events/EnhancedEconomySettingsUpdateEvent.lua')
source(g_currentModDirectory .. 'scripts/events/EnhancedEconomySettingsBroadcastEvent.lua')

local initialised = false
local previousEnhancedEconomySettings = {}

function EnhancedEconomySettings.init()
  if not initialised then
    addModEventListener(EnhancedEconomySettings)

    -- economySettings fillType names
    EnhancedEconomySettings.economySettings = {}

    -- economySettings fillType labels
    EnhancedEconomySettings.economySettingsLabels = {}

    -- Read and write settings from/to XML
    Mission00.loadMission00Finished = Utils.appendedFunction(Mission00.loadMission00Finished,
      EnhancedEconomySettings.loadMission00Finished)
    FSCareerMissionInfo.saveToXMLFile = Utils.appendedFunction(FSCareerMissionInfo.saveToXMLFile,
      EnhancedEconomySettings.saveToXMLFile)

    -- Hook into connection finished loading to send request for settings to clients from server
    FSBaseMission.onConnectionFinishedLoading = Utils.appendedFunction(FSBaseMission.onConnectionFinishedLoading,
      EnhancedEconomySettings.onConnectionFinishedLoading)

    -- Hook into SellingStation.setPriceMultiplier (not required for BuyingStation)
    SellingStation.setPriceMultiplier = Utils.overwrittenFunction(SellingStation.getPriceMultiplier,
      EnhancedEconomySettings.setSellingStationPriceMultiplier)

    EnhancedEconomySettingsGUI.init()

    initialised = true
  end
end

function EnhancedEconomySettings:setSellingStationPriceMultiplier(super, fillTypeIndex, multiplier, ...)
  -- cache the given value
  if (self.originalFillTypePriceMultipliers == nil) then
    self.originalFillTypePriceMultipliers = {}
  end
  if (self.originalFillTypePriceMultipliers[fillTypeIndex] == nil) then
    self.originalFillTypePriceMultipliers[fillTypeIndex] = multiplier
  end

  if (EnhancedEconomySettings.economySettings[fillTypeIndex] ~= nil) then
    -- set multiplied values
    return super(self, fillTypeIndex, multiplier * EnhancedEconomySettings.economySettings[fillTypeIndex], ...)
  end
  -- set given value
  return super(self, fillTypeIndex, multiplier, ...)
end

function EnhancedEconomySettings:update()
  -- for each animal fill type
  for fillTypeIndex, animalSubType in pairs(g_currentMission.animalSystem.fillTypeIndexToSubType) do
    if (animalSubType.buyPrice ~= nil
          and animalSubType.buyPrice.keyframes ~= nil) then
      -- for each key frame
      for keyframeIndex, keyframe in pairs(animalSubType.buyPrice.keyframes) do
        -- for each element of key frame
        for keyframeSubIndex, keyframeValue in pairs(keyframe) do
          -- if the key frame sub index is a number
          if type(keyframeSubIndex) == 'number' then
            if (animalSubType.originalBuyPriceKeyframes == nil) then
              animalSubType.originalBuyPriceKeyframes = {}
            end
            if (animalSubType.originalBuyPriceKeyframes[keyframeIndex] == nil) then
              animalSubType.originalBuyPriceKeyframes[keyframeIndex] = {}
            end
            if (animalSubType.originalBuyPriceKeyframes[keyframeIndex][keyframeSubIndex] == nil) then
              animalSubType.originalBuyPriceKeyframes[keyframeIndex][keyframeSubIndex] = keyframeValue
            end

            -- if a valid economy setting
            if (EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]) then
              -- apply modified key frame value
              animalSubType.buyPrice.keyframes[keyframeIndex][keyframeSubIndex] = animalSubType
              .originalBuyPriceKeyframes[keyframeIndex][keyframeSubIndex] *
              EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]
            end
          end
        end
      end
    end

    if (animalSubType.sellPrice ~= nil
          and animalSubType.sellPrice.keyframes ~= nil) then
      -- for each key frame
      for keyframeIndex, keyframe in pairs(animalSubType.sellPrice.keyframes) do
        -- for each element of key frame
        for keyframeSubIndex, keyframeValue in pairs(keyframe) do
          -- if the key frame sub index is a number
          if type(keyframeSubIndex) == 'number' then
            if (animalSubType.originalSellPriceKeyframes == nil) then
              animalSubType.originalSellPriceKeyframes = {}
            end
            if (animalSubType.originalSellPriceKeyframes[keyframeIndex] == nil) then
              animalSubType.originalSellPriceKeyframes[keyframeIndex] = {}
            end
            if (animalSubType.originalSellPriceKeyframes[keyframeIndex][keyframeSubIndex] == nil) then
              animalSubType.originalSellPriceKeyframes[keyframeIndex][keyframeSubIndex] = keyframeValue
            end

            -- if a valid economy setting
            if (EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]) then
              -- apply modified key frame value
              animalSubType.sellPrice.keyframes[keyframeIndex][keyframeSubIndex] = animalSubType
              .originalSellPriceKeyframes[keyframeIndex][keyframeSubIndex] *
              EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]
            end
          end
        end
      end
    end
  end

  -- for each placeable
  for _, placeable in pairs(g_currentMission.placeableSystem.placeables) do
    if (placeable.typeName == 'buyingStation'
          and placeable.spec_buyingStation ~= nil
          and placeable.spec_buyingStation.buyingStation ~= nil
          and placeable.spec_buyingStation.buyingStation.fillTypePricesScale ~= nil) then
      local station = placeable.spec_buyingStation.buyingStation

      -- for each fillTypePricesScale on buying station
      for fillTypeIndex, pricesScale in pairs(placeable.spec_buyingStation.buyingStation.fillTypePricesScale) do
        -- if not an animal fill type
        if (g_currentMission.animalSystem.fillTypeIndexToSubType[fillTypeIndex] == nil) then
          if (station.originalFillTypePricesScale == nil) then
            station.originalFillTypePricesScale = {}
          end
          if (station.originalFillTypePricesScale[fillTypeIndex] == nil) then
            station.originalFillTypePricesScale[fillTypeIndex] = pricesScale
          end

          -- if a valid economy setting
          if (EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]) then
            -- apply modified prices scale
            station.fillTypePricesScale[fillTypeIndex] = station.originalFillTypePricesScale[fillTypeIndex] *
            EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]
          end
        end
      end
    elseif (placeable.typeName == 'sellingStation' or placeable.typeName == 'productionPoint') then
      local station = nil

      if (placeable.spec_sellingStation ~= nil
            and placeable.spec_sellingStation.sellingStation ~= nil
            and placeable.spec_sellingStation.sellingStation.priceMultipliers ~= nil) then
        station = placeable.spec_sellingStation.sellingStation
      elseif (placeable.spec_productionPoint ~= nil
            and placeable.spec_productionPoint.productionPoint ~= nil
            and placeable.spec_productionPoint.productionPoint.unloadingStation ~= nil
            and placeable.spec_productionPoint.productionPoint.unloadingStation.priceMultipliers ~= nil) then
        station = placeable.spec_productionPoint.productionPoint.unloadingStation
      end

      if (station ~= nil) then
        -- for each priceMultiplier on station
        for fillTypeIndex, priceMultiplier in pairs(station.priceMultipliers) do
          -- if not an animal fill type
          if (g_currentMission.animalSystem.fillTypeIndexToSubType[fillTypeIndex] == nil) then
            if (station.originalFillTypePriceMultipliers == nil) then
              station.originalFillTypePriceMultipliers = {}
            end
            if (station.originalFillTypePriceMultipliers[fillTypeIndex] == nil) then
              station.originalFillTypePriceMultipliers[fillTypeIndex] = priceMultiplier
            end

            -- if a valid economy setting
            if (EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]) then
              -- apply modified price multplier
              station.priceMultipliers[fillTypeIndex] = station.originalFillTypePriceMultipliers[fillTypeIndex] *
              EnhancedEconomySettings.economySettings[g_fillTypeManager.indexToName[fillTypeIndex]]
            end
          end
        end
      end
    end
  end
end

function EnhancedEconomySettings.saveToXMLFile(missionInfo)
  if (EnhancedEconomySettings.economySettings == nil) then
    return
  end

  local xmlFile = XMLFile.create('EnhancedEconomySettings',
    missionInfo.savegameDirectory .. '/EnhancedEconomySettings.xml', 'enhancedEconomySettings')
  if xmlFile ~= nil then
    for fillTypeName, value in pairs(EnhancedEconomySettings.economySettings) do
      xmlFile:setFloat('enhancedEconomySettings.' .. fillTypeName, value)
    end

    xmlFile:save()
    xmlFile:delete()
  end
end

function EnhancedEconomySettings.loadMission00Finished(mission, node)
  -- initialise local enhancedEconomySettings multipliers (on client and server)
  for _, fillType in pairs(g_fillTypeManager.fillTypes) do
    -- If fill type is an animal or if startPricePerLiter not zero
    -- Otherwise value will remain nil to signify it should be disabled in UI
    if (g_currentMission.animalSystem.fillTypeIndexToSubType[fillType.index] ~= nil or fillType.startPricePerLiter > 0) then
      -- initialise to 1 (regular price)
      EnhancedEconomySettings.economySettings[fillType.name] = 1
      previousEnhancedEconomySettings[fillType.name] = 1
    end

    -- store the names to use
    local fillTypeTitle = fillType.title or fillType.name
    local fillTypeName = (fillTypeTitle ~= fillType.name and fillType.name or nil)
    local visualStoreName = nil

    if (g_currentMission.animalSystem.fillTypeIndexToSubType[fillType.index] ~= nil) then
      local visual = nil
      if (g_currentMission.animalSystem.fillTypeIndexToSubType[fillType.index].visuals ~= nil) then
        for _, v in pairs(g_currentMission.animalSystem.fillTypeIndexToSubType[fillType.index].visuals) do
          if ((visual == nil or (v.minAge or 0) >= (visual.minAge or 0)) and v.store ~= nil and v.store.name ~= nil) then
            visual = v
          end
        end

        if (visual ~= nil) then
          visualStoreName = visual.store.name
        end
      end

      if (visualStoreName or fillTypeName) then
        EnhancedEconomySettings.economySettingsLabels[fillType.name] = fillTypeTitle .. ' (' .. (visualStoreName or fillTypeName) .. ')'
      else
        EnhancedEconomySettings.economySettingsLabels[fillType.name] = fillTypeTitle
      end
    else
      EnhancedEconomySettings.economySettingsLabels[fillType.name] = fillTypeTitle
    end

    -- find duplicate labels
    local duplicateLabels = {}
    for _, label1 in pairs(EnhancedEconomySettings.economySettingsLabels) do
      local count = 0
      for _, label2 in pairs(EnhancedEconomySettings.economySettingsLabels) do
        if (label1 == label2) then
          count = count + 1
        end
      end

      if (count > 1) then
        table.insert(duplicateLabels, label1)
      end
    end

    -- attach fillType.name to differentiate duplicate labels
    for _, label in pairs(duplicateLabels) do
      for fillTypeName, originalLabel in pairs(EnhancedEconomySettings.economySettingsLabels) do
        if (label == originalLabel) then
          EnhancedEconomySettings.economySettingsLabels[fillTypeName] = originalLabel .. ' (' .. fillTypeName .. ')'
        end
      end
    end
  end

  if mission:getIsServer() then
    -- only load XML if server
    if mission.missionInfo.savegameDirectory ~= nil and fileExists(mission.missionInfo.savegameDirectory .. '/EnhancedEconomySettings.xml') then
      local xmlFile = XMLFile.load('EnhancedEconomySettings',
        mission.missionInfo.savegameDirectory .. '/EnhancedEconomySettings.xml')
      if xmlFile ~= nil then
        for fillTypeName, value in pairs(EnhancedEconomySettings.economySettings) do
          -- if the setting for this fillType is not disabled
          if (value ~= nil) then
            EnhancedEconomySettings.loadFloatFromXML(xmlFile, 'enhancedEconomySettings.' .. fillTypeName, fillTypeName)
          end
        end

        xmlFile:delete()
      end
    end

    -- update based on local enhancedEconomySettings multipliers
    EnhancedEconomySettings:updateEconomy(EnhancedEconomySettings.economySettings, true)
  end
end

function EnhancedEconomySettings.loadFloatFromXML(xmlFile, xmlKey, fillTypeName)
  local value = xmlFile:getFloat(xmlKey)
  if value == nil or not EnhancedEconomySettings:tableContains(EnhancedEconomySettingsGUI.multiplierOptions, value) then
    return
  end

  EnhancedEconomySettings.economySettings[fillTypeName] = value
end

function EnhancedEconomySettings.onConnectionFinishedLoading()
  EnhancedEconomySettingsBroadcastEvent:sendEvent(EnhancedEconomySettings.economySettings)
end

function EnhancedEconomySettings:updateEconomy(enhancedEconomySettings, hideTopNotification)
  -- for each local fill type
  for fillTypeName, value in pairs(enhancedEconomySettings) do
    -- if new value is different to current
    if (value ~= previousEnhancedEconomySettings[fillTypeName]) then
      previousEnhancedEconomySettings[fillTypeName] = value
      local fillType = g_fillTypeManager.nameToFillType[fillTypeName]

      if (hideTopNotification == true) then
        g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_OK,
          EnhancedEconomySettings.economySettingsLabels[fillType.name] .. ' ' ..
          EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_Notification' ..
          EnhancedEconomySettings:round(enhancedEconomySettings[fillTypeName], 2)))
      else
        g_currentMission.hud:addTopNotification(EnhancedEconomySettings.economySettingsLabels[fillType.name],
          EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_Notification' ..
          EnhancedEconomySettings:round(enhancedEconomySettings[fillTypeName], 2)),
          EnhancedEconomySettingsGUI:getText('EnhancedEconomySettings_Subtitle'), nil, 5000)
      end
    end
  end

  -- update enhancedEconomySettings
  EnhancedEconomySettings.economySettings = enhancedEconomySettings;

  -- update menu in case it is open
  if (EnhancedEconomySettingsGUI.frame ~= nil) then
    EnhancedEconomySettingsGUI.frame.updateGameSettings(EnhancedEconomySettingsGUI.frame)
  end
end

function EnhancedEconomySettings:round(num, dp)
  local m = 10 ^ (dp or 0)
  return math.floor(num * m + 0.5) / m
end

function EnhancedEconomySettings:tableContains(tbl, value)
  for i = 1, #tbl do
    if (tbl[i] == EnhancedEconomySettings:round(value, 2)) then
      return i
    end
  end
  return nil
end

EnhancedEconomySettings.init()
